﻿namespace R2Z.Shields.Printing
{
    using System;
    using System.Drawing;
    using System.IO.Ports;
    using System.Text;


    /// <summary>
    /// An abstraction layer for a PrinterShield connected via USB or generic UART
    /// </summary>
    /// <remarks>R2Z Lab - Written by Luca Rapotec - luca@rapotec.info</remarks>
    public class PrinterShield : IDisposable
    {
        private bool _disposed;
        private SerialPort _serialPort;

        private const byte START_GRAPHIC_MODE = 0x11;
        private const byte PRINT_SERIAL_NUMBER = 0x13;
        private const byte PRINT_DEVICE_INFO = 0x14;

        #region Public constants

        /// <summary>
        /// Height in dots of a printer line
        /// </summary>
        public const int LINE_HEIGHT_DOTS = 7;

        /// <summary>
        /// Width in dots of a printer line
        /// </summary>
        public const int LINE_WIDTH_DOTS = 105;

        /// <summary>
        /// Dead distance in dots betwheen two lines
        /// </summary>
        public const int INTERLINE_DEAD_DOTS = 4;

        /// <summary>
        /// Lenght in characters of a printer line
        /// </summary>
        public const int CHARACTERS_PER_LINE = 15;

        /// <summary>
        /// Single character width in dots
        /// </summary>
        public const int CHARACTER_WIDTH_DOTS = 5;

        /// <summary>
        /// Single character height in dots
        /// </summary>
        public const int CHARACTER_HEIGHT_DOTS = LINE_HEIGHT_DOTS;

        /// <summary>
        /// Dots betwheen characters
        /// </summary>
        public const int CHARACTER_DISTANCE_DOTS = 2;

        #endregion

        #region Public methods

        /// <summary>
        /// Create a new instance of a Printer Shield
        /// </summary>
        /// <param name="comPortNumber">COM port number</param>
        public PrinterShield(string comPort)
        {
            _disposed = false;
            _serialPort = new SerialPort(comPort);
            _serialPort.BaudRate = 115200;
        }

        /// <summary>
        /// Print some text
        /// </summary>
        /// <param name="line"></param>
        public void PrintLine(string line)
        {
            AssertNotDisposed();
            line = line.PadRight(15, ' ');

            WriteByte(0xD); // Do carriage return (reset internal cursor)

            byte[] bLine = Encoding.ASCII.GetBytes(line);

            for (uint x = 0; x < bLine.Length; x++)
            {
                WriteByte(bLine[x]);
            }
        }


        /// <summary>
        /// Print an image from a Bitmap object
        /// </summary>
        /// <remarks>Remember to ALWAYS dispose a Bitmap object!</remarks>
        /// <param name="bmp">A valid Bitmap object</param>
        public void PrintImage(Bitmap bmp)
        {
            AssertNotDisposed();

            int lines = (bmp.Height / (LINE_HEIGHT_DOTS + INTERLINE_DEAD_DOTS));
            if ((bmp.Height % (LINE_HEIGHT_DOTS + INTERLINE_DEAD_DOTS)) > 0)
            {
                lines++;
            }

            for (uint x = 0; x < lines; x++)
            {
                byte[] line = GetBitmapLine(bmp, x);
                StartGraphicMode();
                for (uint k = 0; k < line.Length; k++)
                {
                    WriteByte(line[k]);
                }
            }
        }

        /// <summary>
        /// Print an image from an existing file
        /// </summary>
        /// <param name="imagePath">Path to an existing image file</param>
        public void PrintImage(string imagePath)
        {
            AssertNotDisposed();

            using (Bitmap bmp = (Bitmap)Bitmap.FromFile(imagePath))
            {
                PrintImage(bmp);
            }
        }

        /// <summary>
        /// Print firmware device info
        /// </summary>
        /// <param name="imagePath"></param>
        public void PrintDeviceInfos()
        {
            AssertNotDisposed();

            WriteByte(PRINT_DEVICE_INFO);
        }


        public string GetSerialNumber()
        {
            AssertNotDisposed();

            WriteByte(PRINT_SERIAL_NUMBER);
            byte[] tmp = new byte[15];

            for (int x = 0; x < 15; x++)
            {
                tmp[x] = (byte)_serialPort.ReadByte();
            }
            return Encoding.ASCII.GetString(tmp);
        }

        /// <summary>
        /// Open PrinterShield's communication channel
        /// </summary>
        /// <returns></returns>
        public bool Open()
        {
            AssertNotDisposed();

            if (_serialPort.IsOpen)
            {
                return false;
            }
            try
            {
                _serialPort.Open();
            }
            catch
            {
                return false;
            }
            return _serialPort.IsOpen;
        }

        /// <summary>
        /// Close PrinterShield's communications channel
        /// </summary>
        /// <returns></returns>
        public bool Close()
        {
            AssertNotDisposed();

            if (!_serialPort.IsOpen)
            {
                return false;
            }
            try
            {
                _serialPort.Close();
            }
            catch
            {
                return false;
            }
            return !_serialPort.IsOpen;
        }

        #endregion

        #region Private methods

        private void AssertNotDisposed()
        {
            if (_disposed) { throw new ObjectDisposedException("PrinterShield"); }
        }

        private void StartGraphicMode()
        {
            WriteByte(START_GRAPHIC_MODE);
        }

        private byte WriteByte(byte data)
        {
            _serialPort.Write(new byte[] { data }, 0, 1);
            return (byte)_serialPort.ReadByte();
        }

        private static byte[] GetBitmapLine(Bitmap bmp, uint lineNumber)
        {
            byte data;
            byte[] line = new byte[LINE_WIDTH_DOTS];
            for (uint x = 0; x < line.Length; x++)
            {
                data = 0;
                for (uint y = 0; y < LINE_HEIGHT_DOTS; y++)
                {
                    int imgY = (int)(y + (lineNumber * (LINE_HEIGHT_DOTS + INTERLINE_DEAD_DOTS)));
                    if (imgY <= bmp.Height && x <= bmp.Width)
                    {
                        data = SetBit(data, (byte)((LINE_HEIGHT_DOTS - 1) - y), bmp.GetPixel((int)x, imgY));
                    }
                }
                line[x] = data;
            }
            return line;
        }

        private static byte SetBit(byte data, byte bit, Color value)
        {
            byte blackValue = RGBToGrey(value);

            if (blackValue < 128)
            {
                data = (byte)(data | (1 << bit));
            }
            else
            {
                data = (byte)(data & ~(1 << bit));
            }
            return data;
        }

        private static byte RGBToGrey(Color color)
        {
            return (byte)(((float)color.R + (float)color.G + (float)color.B) / (float)3);
        }

        public void Dispose()
        {
            _disposed = true;
            try { _serialPort.Close(); } catch { }
            try { _serialPort.Dispose(); } catch { }
            GC.SuppressFinalize(this);
        }

        #endregion
    }
}